@@ -1,7 +1,12 @@
 #include "S.h"
 #include "Rinternals.h"
 
+void F77_NAME(adapt)(int *ndim, double *a, double *b, int *minpts, int
+		     *maxpts, double *eps, double *relerr, int *lenwrk,
+		     double *wrkstr, double *finest, int *ifail);
+
 static SEXP rho;
+static SEXP f;
 
 /* All this routine does is call the approriate fortran
    function.  We need this so as to properly pass the S function */
@@ -15,18 +20,20 @@ void cadapt(int *ndim, double *lower, double *upper,
   double *wrkstr;
   wrkstr = (double *) S_alloc(*lenwrk, sizeof(double));
 
+  /* store the R function and its environment */
   rho=env;
+  f=functn;
 
-  F77_CALL(adapt)(ndim,lower,upper,minpts,maxpts,functn,eps,relerr,lenwrk,
+  F77_CALL(adapt)(ndim,lower,upper,minpts,maxpts,eps,relerr,lenwrk,
 		  wrkstr,finest,ifail);
 }
 
 /* This is the fixed routine called by adapt */
 /* changed to double for R, also rewritten to use eval() */
 
-double F77_NAME(adphlp)(void *f, int *ndim, double *z)
+double F77_NAME(adphlp)(int *ndim, double *z)
 {
-  SEXP args,resultsxp;
+  SEXP args,resultsxp,callsxp;
   double result;
   int i;
 
@@ -35,11 +42,12 @@ double F77_NAME(adphlp)(void *f, int *ndim, double *z)
     REAL(args)[i]=z[i];
   }
 
-  PROTECT(resultsxp=eval(lang2( (SEXP) f,args), (SEXP) rho));
+  PROTECT(callsxp=lang2( f,args));
+  PROTECT(resultsxp=eval(callsxp,rho));
 
   result=REAL(resultsxp)[0];
 
-  UNPROTECT(2);
+  UNPROTECT(3);
 
   return(result);
 }
